import json
import pdal
import logging
import os
import subprocess as sp
from typing import Tuple

def las_info_metadata(filename: str):
    """Get las info from pdal info --metadata"""
    ret = sp.run(["pdal", "info", filename, "--metadata"], capture_output=True)
    if ret.returncode == 0:
        infos = ret.stdout.decode()
        infos = json.loads(infos)

        return infos['metadata']

    else:
        raise RuntimeError(f"pdal info failed with error: \n {ret.stderr}")


def las_info_pipeline(filename:str,  spatial_ref:str="EPSG:2154"):
    """Get las info from pdal pipeline with filter.info
    Args:
        filename: input las
        spatial_ref: spatial reference to pass as 'override_srs' argument in las reader
    """
    information = {
    "pipeline": [
            {
                "type": "readers.las",
                "filename": filename,
                "override_srs": spatial_ref,
                "nosrs": True
            },
            {
                "type": "filters.info"
            }
        ]
    }

    # Create json
    json_info = json.dumps(information, sort_keys=True, indent=4)
    logging.info(json_info)
    pipeline = pdal.Pipeline(json_info)
    pipeline.execute()
    pipeline.arrays
    # Extract metadata
    metadata = pipeline.metadata

    if type(metadata) == str:
        metadata = json.loads(metadata)

    return metadata['metadata']['filters.info']


def las_get_xy_bounds(filename: str, buffer_width: int=0, spatial_ref:str="EPSG:2154"):
    """ Get tile bounds (xy only) from las metadata.
    Try getting bounds using las_info_metadata
    As command "pdal_info --metadata" does not seem to work properly on some data
    (TerraSolid output for ex), fallback to las_info_pipeline

    Args:
        filename (str): full path of file for which to get the bounding box
        buffer_width (str): number of pixel to add to the bounding box on each side (buffer size)
        spatial_ref: spatial reference to pass as 'override_srs' argument in las reader
        (Used if pdal info --metadata failed)

    Returns:
        bounds(tuple) : Tuple of bounding box from the LIDAR tile with potential buffer
    """
    # Parameters
    _x = []
    _y = []
    bounds= []
    try:
        metadata = las_info_metadata(filename)
        bounds_dict = metadata

    except RuntimeError as e:
        metadata = las_info_pipeline(filename, spatial_ref)
        bounds_dict = metadata["bbox"]
    if type(metadata) == str:
        metadata = json.loads(metadata)
    # Export bound (maxy, maxy, minx and miny), then creating a buffer with 100 m
    _x.append(float((bounds_dict['minx']) - buffer_width)) # coordinate minX
    _x.append(float((bounds_dict['maxx']) + buffer_width)) # coordinate maxX
    _y.append(float((bounds_dict['miny']) - buffer_width)) # coordinate minY
    _y.append(float((bounds_dict['maxy']) + buffer_width)) # coordinate maxY
    bounds.append(_x) # [xmin, xmax]
    bounds.append(_y) # insert [ymin, ymax]

    return tuple(i for i in bounds)


def parse_filename(file: str):
    """Parse filename and return prefix, suffix and coordinates.
    It expects coordinates to be formatted as {prefix1}_{prefix2}_{coordx}_{coordy}_{suffix}
    For example Semis_2021_0000_1111_LA93_IGN69.las"""
    basename = os.path.basename(file)  # Make sure that we work on the base name and not the full path

    prefix1, prefix2, coordx, coordy, suffix = basename.split("_", 4)
    prefix = f"{prefix1}_{prefix2}"

    return prefix, int(coordx), int(coordy), suffix


def get_buffered_bounds_from_filename(filename: str, buffer_width: int=0,
                                      tile_width: int=1000, tile_coord_scale: int=1000) -> Tuple:
    """ Get tile bounds (xy only) from las metadata.
    Try getting bounds using las_info_metadata
    As command "pdal_info --metadata" does not seem to work properly on some data
    (TerraSolid output for ex), fallback to las_info_pipeline

    Args:
        filename (str): full path of file for which to get the bounding box
        buffer_width (str): number of pixel to add to the bounding box on each side (buffer size)
        tile_width (int): Width of a tile(in the reference unit: 1m)
        tile_coord_scale (int): Scale used in filename to describe coordinates (usually kilometers)
            1000 * 1m (with 1m being the reference)

    Returns:
        bounds(tuple) : Tuple of bounding box from the LIDAR tile with potential buffer
    """
    _, coordX, coordY, _ = parse_filename(filename)
    # Coordinates in the filenames are x_min and y_max
    minX = coordX * tile_coord_scale
    maxX = coordX * tile_coord_scale + tile_width
    maxY = coordY * tile_coord_scale
    minY = coordY * tile_coord_scale - tile_width

    xs = [minX - buffer_width, maxX + buffer_width]
    ys = [minY - buffer_width, maxY + buffer_width]

    return (xs,ys)


import laspy
import os
import shutil
import pytest

from pdaltools.color import color, pdal_info_json
from pdaltools.unlock_file import unlock_file, copy_and_hack_decorator


TEST_PATH = os.path.dirname(os.path.abspath(__file__))
TMPDIR = os.path.join(TEST_PATH, "tmp")
LAZ_FILE = os.path.join(TEST_PATH, 'data/test_pdalfail_0643_6319_LA93_IGN69.laz')


def setup_module(module):
    try:
        shutil.rmtree(TMPDIR)
    except (FileNotFoundError):
        pass
    os.mkdir(TMPDIR)


@copy_and_hack_decorator
def decorated_pdal_info_json(input_file: str):
    return pdal_info_json(input_file)


def test_copy_and_hack_decorator_simple():
    TMP_FILE = os.path.join(TMPDIR, "copy_and_hack_simple.laz")
    shutil.copy(LAZ_FILE, TMP_FILE)
    decorated_pdal_info_json(TMP_FILE)


@pytest.mark.geoportail
def test_copy_and_hack_decorator_color():
    # bug during laz opening in pdal (solved with copy_and_hack_decorator)
    LAS_FILE = os.path.join(TMPDIR, "test_pdalfail_0643_6319_LA93_IGN69.las")

    color(LAZ_FILE, LAS_FILE, "", 1)

    las = laspy.read(LAS_FILE)
    print(las.header)
    print(list(las.point_format.dimension_names))
    print(las.red)
    print(las.green)
    print(las.blue)
    print(las.nir)

    assert os.path.isfile(LAS_FILE)


def test_unlock_file():
    TMP_FILE = os.path.join(TMPDIR, "unlock_file.laz")
    shutil.copy(LAZ_FILE, TMP_FILE)
    unlock_file(TMP_FILE)
    pdal_info_json(TMP_FILE)


from collections import Counter
import json
import os
from pdaltools import las_info
import pytest
import shutil
from test.utils import get_pdal_infos_summary
from typing import Dict
import numpy as np


coord_x = 77055
coord_y = 627760
tile_width = 50
tile_coord_scale = 10
test_path = os.path.dirname(os.path.abspath(__file__))
tmp_path = os.path.join(test_path, "tmp")
input_dir = os.path.join(test_path, "data")
input_file = os.path.join(input_dir, f"test_data_{coord_x}_{coord_y}_LA93_IGN69_ground.las")
input_mins = [ 770550., 6277550.]
input_maxs = [ 770600., 6277600.]


def test_las_info_metadata():
    metadata = las_info.las_info_metadata(input_file)
    assert type(metadata) == dict
    assert bool(metadata)  # check that metadata dict is not empty


def test_las_info_pipeline():
    info = las_info.las_info_pipeline(input_file)
    assert type(info) == dict
    assert bool(info)  # check that info dict is not empty


def test_las_get_xy_bounds_no_buffer():
    bounds = las_info.las_get_xy_bounds(input_file)
    expected_xs = [input_mins[0], input_maxs[0]]
    expected_ys = [input_mins[1], input_maxs[1]]
    assert np.allclose(bounds, [expected_xs, expected_ys], rtol=1e-06)


def test_las_get_xy_bounds_with_buffer():
    buffer_width = 10
    bounds = las_info.las_get_xy_bounds(input_file, buffer_width=buffer_width)
    expected_xs = [input_mins[0] - buffer_width, input_maxs[0] + buffer_width]
    expected_ys = [input_mins[1] - buffer_width, input_maxs[1] + buffer_width]
    assert np.allclose(bounds, [expected_xs, expected_ys], rtol=1e-06)


def test_parse_filename():
    prefix, parsed_coord_x, parsed_coord_y, suffix = las_info.parse_filename(input_file)
    assert prefix == "test_data"
    assert suffix == "LA93_IGN69_ground.las"
    assert parsed_coord_x == coord_x
    assert parsed_coord_y == coord_y


def test_get_buffered_bounds_from_filename_no_buffer():
    xs, ys = las_info.get_buffered_bounds_from_filename(input_file, tile_width=tile_width,
                                                        tile_coord_scale=tile_coord_scale)
    assert xs == [770550, 770600]
    assert ys == [6277550, 6277600]


def test_get_buffered_bounds_from_filename_with_buffer():
    buffer_width = 10
    xs, ys = las_info.get_buffered_bounds_from_filename(input_file, tile_width=tile_width,
                                                        tile_coord_scale=tile_coord_scale,
                                                        buffer_width=buffer_width)
    assert xs == [770550 - buffer_width, 770600 + buffer_width]
    assert ys == [6277550 - buffer_width, 6277600 + buffer_width]


# -*- coding: utf-8 -*-
# maintener : MDupays
# version : v.1 06/12/2022
# Extract info from the tile
import logging
import pdal
import json


def las_crop(input_file: str, output_file: str, bounds, spatial_ref:str="EPSG:2154"):
    """ Crop filter removes points that fall inside a cropping bounding box (2D)
    Args:
        input_dir (str): input point cloud file
        output_dir (str): output point cloud file
        bounds : 2D bounding box to crop to : provided as ([xmin, xmax], [ymin, ymax])
    """
    # Parameters
    information = {
    "pipeline": [
            {
                "type": "readers.las",
                "filename": input_file,
                "override_srs": spatial_ref,
                "nosrs": True
            },
            {
                "type":"filters.crop",
                "bounds": str(bounds)
            },
            {
                "type": "writers.las",
                "a_srs": spatial_ref,
                "filename": output_file
            }
        ]
    }
    # Create json
    json_crop = json.dumps(information, sort_keys=True, indent=4)
    logging.info(json_crop)
    pipeline = pdal.Pipeline(json_crop)
    pipeline.execute()



import os
import shutil
import pytest

from pdaltools import color

import requests
import requests_mock

cwd = os.getcwd()

TMPDIR = cwd + "/tmp/"

def setup_module(module):
    try:
        shutil.rmtree(TMPDIR)
    except (FileNotFoundError):
        pass
    os.mkdir(TMPDIR)

TEST_PATH = os.path.dirname(os.path.abspath(__file__))
INPUT_PATH = os.path.join(TEST_PATH, 'data/test_noepsg_043500_629205_IGN69.laz')

OUTPUT_FILE = TMPDIR + "Semis_2021_0435_6292_LA93_IGN69.las"


@pytest.mark.geoportail
def test_epsg_fail():
    with pytest.raises(requests.exceptions.HTTPError, match="400 Client Error: BadRequest for url") :
        color.color(INPUT_PATH, OUTPUT_FILE, "", 0.1, 15)


epsg = "2154"
layer= "ORTHOIMAGERY.ORTHOPHOTOS"
minx=435000
miny=6291000
maxx=436000
maxy=6292000
pixel_per_meter=0.1


@pytest.mark.geoportail
def test_download_image_ok():
    color.download_image_from_geoportail(epsg, layer, minx, miny, maxx, maxy, pixel_per_meter, OUTPUT_FILE, 15)


@pytest.mark.geoportail
def test_download_image_raise1():
    retry_download = color.retry(2, 5)(color.download_image_from_geoportail)
    with pytest.raises(requests.exceptions.HTTPError):
        retry_download(epsg, "MAUVAISE_COUCHE", minx, miny, maxx, maxy, pixel_per_meter, OUTPUT_FILE, 15)


@pytest.mark.geoportail
def test_download_image_raise2():
    retry_download = color.retry(2, 5)(color.download_image_from_geoportail)
    with pytest.raises(requests.exceptions.HTTPError):
        retry_download("9001", layer, minx, miny, maxx, maxy, pixel_per_meter, OUTPUT_FILE, 15)


def test_retry_on_server_error():
    with requests_mock.Mocker() as mock:
        mock.get(requests_mock.ANY, status_code=502, reason="Bad Gateway")
        with pytest.raises(requests.exceptions.HTTPError):
            retry_download = color.retry(2, 1, 2)(color.download_image_from_geoportail)
            retry_download(epsg, layer, minx, miny, maxx, maxy, pixel_per_meter, OUTPUT_FILE, 15)
        history = mock.request_history
        assert len(history) == 3


def test_retry_on_connection_error():
      with requests_mock.Mocker() as mock:
        mock.get(requests_mock.ANY, exc=requests.exceptions.ConnectionError)
        with pytest.raises(requests.exceptions.ConnectionError):
            retry_download = color.retry(2, 1)(color.download_image_from_geoportail)
            retry_download(epsg, layer, minx, miny, maxx, maxy, pixel_per_meter, OUTPUT_FILE, 15)

        history = mock.request_history
        assert len(history) == 3


def test_retry_param():

    # Here you can change retry params
    @color.retry(7, 15, 2, True)
    def raise_server_error():
        raise requests.exceptions.HTTPError("Server Error")

    with pytest.raises(requests.exceptions.HTTPError):
        raise_server_error()



from collections import Counter
import json
import os
from pdaltools.replace_attribute_in_las import replace_values, replace_values_clean, parse_replacement_map_from_path_or_json_string
from pdaltools.count_occurences.count_occurences_for_attribute import compute_count_one_file
from pdaltools.standardize_format import get_writer_parameters
import pytest
import shutil
from test.utils import get_pdal_infos_summary
from typing import Dict
from test.test_standardize_format import assert_lasinfo_no_warning

test_path = os.path.dirname(os.path.abspath(__file__))
tmp_path = os.path.join(test_path, "tmp")
input_dir = os.path.join(test_path, "data/classified_laz")
input_file = os.path.join(input_dir, "test_data_77050_627755_LA93_IGN69.laz")
output_file = os.path.join(tmp_path, "replaced.las")
attribute = "Classification"
input_counts = Counter({
    '1': 2047,
    '2': 21172,
    '3': 226,
    '4': 1227,
    '5': 30392,
    '6': 29447,
    '64': 13,
})

expected_counts = Counter({
    '2': 21172,
    '3': 226,
    '4': 1227,
    '5': 30392,
    '64': 29447,
    '201': 2047 + 13
})

replacement_map_fail = {
    "201" : ["1", "64"],
    "6": ["64"],
}  # has duplicatevalue to replace

replacement_map_success = {
    "201" : ["1", "64"],
    "64": ["6"],
}

# test replacement map parsing
input_parsing_correct = '{ "201" : ["1", "64"],  "64": ["6"]}'
input_parsing_incorrect = '"201" : ["1", "64"],  "64": ["6"]'
input_parsing_file = os.path.join(test_path, "data", "example_replacement_map.json")


def setup_module(module):
    try:
        shutil.rmtree(tmp_path)

    except (FileNotFoundError):
        pass
    os.mkdir(tmp_path)


def test_replace_values():
    replace_values(input_file, output_file, replacement_map_success, attribute)
    count = compute_count_one_file(output_file, attribute)

    assert count == expected_counts
    check_dimensions(input_file, output_file)


def test_replace_values_clean():
    replace_values_clean(input_file, output_file, replacement_map_success, attribute, get_writer_parameters({}))
    assert_lasinfo_no_warning(output_file)


def test_replace_values_duplicate_input():
    with pytest.raises(ValueError):
        replace_values(input_file, output_file, replacement_map_fail, attribute)


def check_dimensions(input_file, output_file):
    input_summary = get_pdal_infos_summary(input_file)
    input_dimensions = set(input_summary["summary"]["dimensions"])
    output_summary = get_pdal_infos_summary(output_file)
    output_dimensions = set(output_summary["summary"]["dimensions"])
    assert input_dimensions == output_dimensions


def test_parse_replacement_map_from_path_or_json_string_path_ok():
    assert isinstance(parse_replacement_map_from_path_or_json_string(input_parsing_file), Dict)


def test_parse_replacement_map_from_path_or_json_string_json_ok():
    assert isinstance(parse_replacement_map_from_path_or_json_string(input_parsing_correct), Dict)


def test_parse_replacement_map_from_path_or_json_string_other_nok():
    with pytest.raises(json.decoder.JSONDecodeError):
        parse_replacement_map_from_path_or_json_string(input_parsing_incorrect)



import json
import subprocess as sp
import tempfile
import pdal
import requests
from osgeo.osr import SpatialReference
import time
import argparse

from pdaltools.unlock_file import copy_and_hack_decorator


def pretty_time_delta(seconds):
    sign_string = '-' if seconds < 0 else ''
    seconds = abs(int(seconds))
    days, seconds = divmod(seconds, 86400)
    hours, seconds = divmod(seconds, 3600)
    minutes, seconds = divmod(seconds, 60)
    if days > 0:
        return '%s%dd%dh%dm%ds' % (sign_string, days, hours, minutes, seconds)
    elif hours > 0:
        return '%s%dh%dm%ds' % (sign_string, hours, minutes, seconds)
    elif minutes > 0:
        return '%s%dm%ds' % (sign_string, minutes, seconds)
    else:
        return '%s%ds' % (sign_string, seconds)


def retry(times, delay, factor=2, debug=False):
    def decorator(func):
        def newfn(*args, **kwargs):
            attempt = 1
            new_delay = delay
            while attempt <= times:
                need_retry = False
                try:
                    return func(*args, **kwargs)
                except requests.exceptions.ConnectionError as err:
                    print ("Connection Error:", err)
                    need_retry = True
                except requests.exceptions.HTTPError as err:
                    if "Server Error" in str(err):
                        print ("HTTP Error:", err)
                        need_retry = True
                    else:
                        raise err
                if need_retry:
                    print(f"{attempt}/{times} Nouvel essai après une pause de {pretty_time_delta(new_delay)} .. ")
                    if not debug:
                        time.sleep(new_delay)
                    new_delay = new_delay * factor
                    attempt += 1

            return func(*args, **kwargs)
        return newfn
    return decorator


def download_image_from_geoportail(
    proj, layer, minx, miny, maxx, maxy, pixel_per_meter, outfile, timeout
):
    # for layer in layers:
    URL_GPP = "https://wxs.ign.fr/ortho/geoportail/r/wms?"
    URL_FORMAT = "&EXCEPTIONS=text/xml&FORMAT=image/geotiff&SERVICE=WMS&VERSION=1.3.0&REQUEST=GetMap&STYLES="
    URL_EPSG = "&CRS=EPSG:" + str(proj)
    URL_BBOX = (
        "&BBOX=" + str(minx) + "," + str(miny) + "," + str(maxx) + "," + str(maxy)
    )
    URL_SIZE = (
        "&WIDTH="
        + str(int((maxx - minx) * pixel_per_meter))
        + "&HEIGHT="
        + str(int((maxy - miny) * pixel_per_meter))
    )

    URL = URL_GPP + "LAYERS=" + layer + URL_FORMAT + URL_EPSG + URL_BBOX + URL_SIZE

    print(URL)
    if timeout < 10:
        print(f"Mode debug avec un timeout à {timeout} secondes")

    req = requests.get(URL, allow_redirects=True, timeout=timeout)
    req.raise_for_status()
    print(f"Ecriture du fichier: {outfile}")
    open(outfile, "wb").write(req.content)


def proj_from_metadata(metadata):
    spatial_wkt = metadata["comp_spatialreference"]
    osr_crs = SpatialReference()
    osr_crs.ImportFromWkt(spatial_wkt)
    authority = osr_crs.GetAttrValue("AUTHORITY", 0)
    if authority == "EPSG":
        proj = osr_crs.GetAttrValue("AUTHORITY", 1)
    else:
        proj = "2154"  # par defaut
    return proj


def pdal_info_json(input_file: str):
    r = sp.run(["pdal", "info", "--metadata", input_file], stderr=sp.PIPE, stdout=sp.PIPE)
    if r.returncode == 1:
        msg = r.stderr.decode()
        print(msg)
        raise RuntimeError(msg)

    output = r.stdout.decode()
    json_info = {}
    try:
        json_info = json.loads(output)
    except:
        print(r.stderr.decode())
        raise
    return json_info


@copy_and_hack_decorator
def color(input_file: str, output_file :str,
    proj="", pixel_per_meter=5, timeout_second=300,
    color_rvb_enabled=True, color_ir_enabled=True, veget_index_file=""
    ):

    json_info = pdal_info_json(input_file)
    metadata = json_info["metadata"]
    minx, maxx, miny, maxy = metadata["minx"], metadata["maxx"], metadata["miny"], metadata["maxy"]

    if proj == "":
        proj = proj_from_metadata(metadata)

    pipeline = pdal.Reader.las(filename=input_file)

    writer_extra_dims = "all"

    # apply decorator to retry 3 times, and wait 30 seconds each times
    download_image_from_geoportail_retrying = retry(7, 15, 2)(download_image_from_geoportail)

    if veget_index_file and veget_index_file != "":
        print(f"Remplissage du champ Deviation à partir du fichier {veget_index_file}")
        pipeline |= pdal.Filter.colorization(raster=veget_index_file, dimensions="Deviation:1:256.0")
        writer_extra_dims = ["Deviation=ushort"]

    if color_rvb_enabled:
        tmp_ortho = tempfile.NamedTemporaryFile().name
        download_image_from_geoportail_retrying(proj, "ORTHOIMAGERY.ORTHOPHOTOS", minx, miny, maxx, maxy, pixel_per_meter, tmp_ortho, timeout_second)
        pipeline|= pdal.Filter.colorization(raster=tmp_ortho, dimensions="Red:1:256.0, Green:2:256.0, Blue:3:256.0")

    if color_ir_enabled:
        tmp_ortho_irc = tempfile.NamedTemporaryFile().name
        download_image_from_geoportail_retrying(proj, "ORTHOIMAGERY.ORTHOPHOTOS.IRC", minx, miny, maxx, maxy, pixel_per_meter, tmp_ortho_irc, timeout_second)
        pipeline |= pdal.Filter.colorization(raster=tmp_ortho_irc, dimensions="Infrared:1:256.0")

    pipeline |= pdal.Writer.las(filename=output_file, extra_dims=writer_extra_dims, minor_version="4", dataformat_id="8")

    print("Traitement du nuage de point")
    pipeline.execute()

    # os.remove(tmp_ortho)
    # os.remove(tmp_ortho_irc)


def parse_args():
    parser = argparse.ArgumentParser("Colorize tool")
    parser.add_argument(
        "--input", "-i",
        type=str,
        required=True,
        help="Input file")
    parser.add_argument(
        "--output", "-o",
        type=str,
        default="",
        help="Output file")
    parser.add_argument(
        "--proj", "-p",
        type=str,
        default = "",
        help="Projection, default will use projection from metadata input")
    parser.add_argument(
        "--resolution", "-r",
        type=float,
        default = 5,
        help="Resolution, in pixel per meter")
    parser.add_argument(
        "--timeout", "-t",
        type=int,
        default = 300,
        help="Timeout, in seconds")
    parser.add_argument('--rvb', action='store_true', help="Colorize RVB")
    parser.add_argument('--ir', action='store_true', help="Colorize IR")
    parser.add_argument(
        "--vegetation",
        type=str,
        default = "",
        help="Vegetation file, value will be stored in Deviation field")
    return  parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    color(args.input, args.output, args.proj, args.resolution, args.timeout, args.rvb, args.ir, args.vegetation)